/* This program simulates several different load balancing  */
/* strategies for Zoom servers (e.g., RANDOM, ROUND_ROBIN). */
/* It reads a 7-column input file of Zoom meeting info, as  */
/* generated from makeworkload.c when the MEETINGS_ONLY and */
/* DEBUG options are both enabled. The output format is     */
/* time series data of load balancing metrics for plotting. */
/*                                                          */
/* Usage: cc -o balance balance.c -lm                       */
/*        ./balance < meetings.dat > stats.dat              */
/*                                                          */
/* Written by Carey Williamson              June 14, 2022   */

#include <stdio.h>
#include <stdlib.h>

#define NUM_MEETINGS 4000
#define NUM_SESSIONS 4000
#define NUM_CONNS 4000
#define INFINITY 9999999.9

#define NUM_SERVERS 20

/* Policies for scheduling and load balancing */
//#define SINGLE_SERVER 1
#define RANDOM 1
//#define ROUND_ROBIN 1
//#define FEWEST_MEETINGS 1
//#define FEWEST_USERS 1
//#define MOST_MEETINGS 1
//#define MOST_USERS 1

/* Verbose debugging, including start/end events and input reading */
//#define DEBUG 1

int main()
  {

    float starts[NUM_MEETINGS], ends[NUM_MEETINGS];
    int participants[NUM_MEETINGS], servers[NUM_MEETINGS];
    int meetings[NUM_SERVERS], users[NUM_SERVERS];
    int minmtg, maxmtg, minusers, maxusers;
    float time, dur, nextstart, nextend;
    int i, j, num, conns, server, startindex, endindex;
    int meetingID, connID, count, done;
    char src, dst;
    
    num = 0;
    while ( scanf("%f %d %d %d %c %c %f", &time, &meetingID, &connID, &count, &src, &dst, &dur) != EOF)
      { 
	starts[num] = time;
	ends[num] = time + dur;
	participants[num] = count;
	servers[num] = -1;
#ifdef DEBUG
	printf("Meeting %d with %d users from %f to %f\n",
		meetingID, count, starts[num], ends[num]);
#endif
	num++; 
      }
    starts[num] = INFINITY;
    ends[num] = INFINITY;
    
    fprintf(stderr, "Read in %d lines of input\n", num);

#ifdef SINGLE_SERVER
    printf("SINGLE SERVER\n");
#endif
#ifdef RANDOM
    printf("RANDOM\n");
#endif
#ifdef ROUND_ROBIN
    printf("ROUND ROBIN\n");
#endif
#ifdef FEWEST_MEETINGS
    printf("FEWEST MEETINGS\n");
#endif
#ifdef MOST_MEETINGS
    printf("MOST MEETINGS\n");
#endif
#ifdef FEWEST_USERS
    printf("FEWEST USERS\n");
#endif
#ifdef MOST_USERS
    printf("MOST USERS\n");
#endif
    /* initialize servers to empty */
    for( i = 0; i < NUM_SERVERS; i++ )
      {
	meetings[i] = 0;
	users[i] = 0;
      }
    done = 0;

loopit:
    
    /* find the next meeting to start */
    startindex = -1;
    nextstart = INFINITY;
    for( i = 0; i < num; i++ )
      {
	if( starts[i] < nextstart )
	  {
	    nextstart = starts[i];
	    startindex = i;
	  }
      }
#ifdef DEBUG
    printf("Next meeting to start is %d with %d users at time %f\n",
	    startindex, participants[startindex], starts[startindex]);
#endif

    /* find the next meeting to end */
    endindex = -1;
    nextend = INFINITY;
    for( i = 0; i < num; i++ )
      {
	if( ends[i] < nextend )
	  {
	    nextend = ends[i];
	    endindex = i;
	  }
      }
#ifdef DEBUG
    printf("Next meeting to end is %d with %d users at time %f\n",
	    endindex, participants[endindex], ends[endindex]);
#endif

    if( nextstart < nextend )
      {
	time = nextstart;
	starts[startindex] = INFINITY;

	/* choose a server */
#ifdef SINGLE_SERVER
	server = 0;
#endif
#ifdef ROUND_ROBIN
	server = startindex % NUM_SERVERS;
#endif
#ifdef RANDOM
	server = random() % NUM_SERVERS;
#endif

	/* check meeting and user counts */
	minmtg = NUM_CONNS;
	minusers = NUM_CONNS;
	maxmtg = 0;
	maxusers = 0;
	for( i = 0; i < NUM_SERVERS; i++ )
	  {
	    if( meetings[i] < minmtg )
	      {
		minmtg = meetings[i];
#ifdef FEWEST_MEETINGS
		server = i;
#endif
	      }
	    if( meetings[i] > maxmtg )
	      {
		maxmtg = meetings[i];
#ifdef MOST_MEETINGS
		server = i;
#endif
	      }
	    if( users[i] < minusers )
	      {
		minusers = users[i];
#ifdef FEWEST_USERS
		server = i;
#endif
	      }
	    if( users[i] > maxusers )
	      {
		maxusers = users[i];
#ifdef MOST_USERS
		server = i;
#endif
	      }
	  }
	printf("Time %f minmtg %d maxmtg %d minusers %d maxusers %d\n",
	       time, minmtg, maxmtg, minusers, maxusers);

	/* assign to the chosen server */
	servers[startindex] = server;
	
	/* update server load */
	meetings[server]++;
	users[server] += participants[startindex];
#ifdef DEBUG
	printf("Added meeting %d with %d users to server %d at time %f\n",
		startindex, participants[startindex], server, time);
	printf("Server %d now has %d meetings and %d users\n",
		server, meetings[server], users[server]);
#endif
      }
    else
      {
	time = nextend;
	ends[endindex] = INFINITY;

	/* find the server hosting this meeting */
	server = servers[endindex];
	
	/* update server load */
	meetings[server]--;
	users[server] -= participants[endindex];
#ifdef DEBUG
	printf("Removed meeting %d with %d users from server %d at time %f\n",
		endindex, participants[endindex], server, time);
	printf("Server %d now has %d meetings and %d users\n",
		server, meetings[server], users[server]);
#endif

	done++;

	/* check meeting and user counts */
	minmtg = NUM_CONNS;
	minusers = NUM_CONNS;
	maxmtg = 0;
	maxusers = 0;
	for( i = 0; i < NUM_SERVERS; i++ )
	  {
	    if( meetings[i] < minmtg )
	      minmtg = meetings[i];
	    if( meetings[i] > maxmtg )
	      maxmtg = meetings[i];
	    if( users[i] < minusers )
	      minusers = users[i];
	    if( users[i] > maxusers )
	      maxusers = users[i];
	  }
	printf("time %f minmtg %d maxmtg %d minusers %d maxusers %d\n",
	       time, minmtg, maxmtg, minusers, maxusers);
      }

    if( done < num )
      goto loopit;
    else printf("The End!\n");
  }
