/* This program generates synthetic Zoom call workloads.    */
/*                                                          */
/* Usage: cc -o makeworkload makeworkload.c -lm             */
/*        ./makeworkload                                    */
/*                                                          */
/* Written by Carey Williamson              June 12, 2022   */

/* Header files */
#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* use "man 3 log" for the math functions */
#include <time.h> /* for optional seeding of PRNG */

/* Manifest constants */
#define NUM_VALUES 3600 /* number of Zoom meetings to generate */
#define MEAN_IAT 60.0  /* seconds */
#define SHORT_IAT 2.2  /* seconds */
#define CLASS_DURATION 3600.0  /* seconds */
#define MIN_DURATION 60.0  /* seconds */
#define MEAN_DURATION 1800.0  /* seconds */
#define MEAN_SIZE 10.0  /* participants */
#define P2P_PROB 0.2  /* P2P Zoom */
#define CLASS_PROB 0.9  /* Zoom class */
#define HOME_PROB 0.5  /* at home (not on campus) */
#define WIFI_PROB 0.8  /* WiFi instead of Ethernet */
#define VANCOUVER_PROB 0.5  /* Zoom server in Vancouver */
#define TORONTO_PROB 0.4  /* Zoom server in Toronto */

// Default setting is for user session level generation (conns).
// Note that sessions sometimes start prior to the meeting host starting,
// so you likely need to sort your workload file into chronological order.
// If you want just the meetings level, then uncomment the next line.
//#define MEETINGS_ONLY 1

/* Full debugging option for each user session of each meeting, */
/* useful when doing detailed statistical analysis of sessions. */
/* Comment out this line to make workload files for the routing */
/* simulation, but uncomment it for load balancing simulation. */
//#define DEBUG 1

/***********************************************************************/
/*                 RANDOM NUMBER GENERATION STUFF                      */
/***********************************************************************/

/* Parameters for random number generation. */
#define MAX_INT 2147483647       /* Maximum positive integer 2^31 - 1 */

/* array used for initializing seeds for Normal distribution */
unsigned short Nvar1_seed[3] = {4,7,2};
unsigned short Nvar2_seed[3] = {5,8,1};
unsigned short Nvar3_seed[3] = {9,6,3};

/* Generate a random floating point number uniformly distributed in [0,1] */
float Uniform01()
  {
    float randnum;
    /* get a random positive integer from random() */
    randnum = (float) 1.0 * random();
    /* divide by max int to get something in between 0.0 and 1.0 */
    randnum = (float) randnum / (1.0 * MAX_INT); 
    return( randnum );
  }


/* Generate a random floating point number from an exponential    */
/* distribution with mean mu.                                     */
float Exponential(mu)
    float mu;
 {
    float randnum, ans;

    /* The algorithm here is simple. Start with a U(0,1) random */
    /* number, and then inverse map it through the cumulative   */
    /* distribution function (CDF) for the exponential distn.   */
    randnum = Uniform01();
    ans = -(mu) * log(randnum);
    return( ans );
  }


/* Generate random positive integer geometrically distributed with mean m */
int Geometric(m)
    double m;
  {
    double p;    /* p is the probability of "success" for each trial */
    int k;

    k = 1;
    p = 1.0 / m;                          /* the inverse of the mean */
    while( Uniform01() > p )
      k++;
    return( k );
  }


/* the rejection method is used for normal distribution 
   (a) Generate two uniform U(0,1) variates u1 and u2. 
   (b) Let x = -log(u1).
   (c) If u2 > e to the power -(x - 1)square by 2, go back to step
   [a].
   (d) Generate u3.
   (e) If u3 > 0.5, return (mean + (std * x)); otherwise return
   (mean - (std * x)). */

double Normal(float mean, float std)
{
  double var_u1, var_u2, var_u3;
  double x, tmp_val;
  
  var_u1 = erand48(Nvar1_seed);
  var_u2 = erand48(Nvar2_seed);
  x = -log(var_u1);
  tmp_val = exp(-((x - 1)*(x - 1)) / 2);
  while(var_u2 > tmp_val)
    {
      var_u1 = erand48(Nvar1_seed);
      var_u2 = erand48(Nvar2_seed);
      x = -log(var_u1);
      tmp_val = exp(-((x - 1)*(x - 1)) / 2);
    }
  var_u3 = erand48(Nvar3_seed);
  if(var_u3 > 0.5)
    return(mean + (std * x));
  else 
    return(mean - (std * x));
}


/***********************************************************************/
/*                 MAIN PROGRAM                                        */
/***********************************************************************/
int main(int argc, char *argv[])
  {
    int i, j, conns, size, flag;
    float uni, iat, dur, timer, starttime, endtime;
    char source, dest;

    /* Set initial seed for random number generator.        */
    srandom(time(NULL));   /* based on time of day */
    srandom(12753);        /* fixed seed for initial testing */

    timer = 0.0;
    conns = 0;
    
    for(i = 0; i < NUM_VALUES; i++)
      {
	/* Determine if this is a peer-to-peer Zoom meeting or not */
	uni = Uniform01();
	if( uni < P2P_PROB )
	  size = 2;
	else size = Geometric(MEAN_SIZE) + 2;

	if( size == 2 )  /* peer-to-peer Zoom */
	  {
	    /* Generate duration from an exponential distribution */
	    dur = Exponential(MEAN_DURATION);
	    while( dur < MIN_DURATION )
	      dur = Exponential(MEAN_DURATION);

	    /* Determine if first user is at home or on campus */
	    uni = Uniform01();
	    if( uni < HOME_PROB )
	      source = 'X';
	    else source = 'P';

	    /* determine if second user is at home or on campus */
	    uni = Uniform01();
	    if( uni < HOME_PROB )
	      dest = 'Y';
	    else dest = 'Q';

	    /* print out the call info */
#ifdef DEBUG
	    /* full debug info, including meeting ID and participants */
	    printf("%8.6f %d %d %d %c %c %8.6f\n",
		   timer, i, conns, size, source, dest, dur);
#else	
	    /* short version, ready to be used for workload.dat file */
	    printf("%8.6f %c %c %8.6f\n", timer, source, dest, dur);
#endif
	    conns++;
	  }
	else
	  {
	    /* Generate duration for meeting */
	    uni = Uniform01();
	    if( uni < CLASS_PROB )
	      dur = CLASS_DURATION + Normal(0,300.0);
	    else dur = Exponential(MEAN_DURATION);
	    endtime = timer + dur;

	    /* determine location of Zoom server */
	    uni = Uniform01();
	    if( uni < VANCOUVER_PROB )
	      dest = 'V';
	    else if( uni < VANCOUVER_PROB + TORONTO_PROB )
	      dest = 'T';
	    else dest = 'O';

	    /* generate the conns for all the users in this Zoom meeting */
	    flag = 0;
	    while( size > 0 )
	      {
		/* determine if user is at home or on campus */
		uni = Uniform01();
		if( uni < HOME_PROB )
		  source = 'H';
		else
		  {
		    /* determine if user is on WiFi or Ethernet */
		    uni = Uniform01();
		    if( uni < WIFI_PROB )
		      source = 'W';
		    else source = 'E';
		  }

		/* adjust start time and duration for non-host participants */
		if( flag == 0 )
		  starttime = timer;
		else starttime = timer + Normal(0.0, 60.0);
		flag++;

		/* print out the Zoom call info */
#ifdef DEBUG
		/* full debug info, including meeting ID and conn ID */
		printf("%8.6f %d %d %d %c %c %8.6f\n",
		       starttime, i, conns, size, source, dest, endtime - starttime);
#else	
		/* short version, ready to be used for workload.dat file */
		printf("%8.6f %c %c %8.6f\n", starttime, source, dest, endtime - starttime);
#endif
		conns++;
		size--;
#ifdef MEETINGS_ONLY
		size = 0;  /* hack to exit loop (i.e., meetings only, not conns) */
#endif
	      }
	  }

	/* Generate inter-arrival time from an exponential distribution */
	/* with time-varying rate based on time of day. */
	if( timer < 28000.0 )
	  iat = Exponential(MEAN_IAT*5.0); /* very light background load */
	else if( timer > 64000.0 )
	  iat = Exponential(MEAN_IAT); /* light background load */
	else if( (int)timer%3600 > 3200 )
	  iat = Exponential(SHORT_IAT);  /* spikes at class start times 8am to 5pm */
	else if( (int)timer%1800 > 1500 )
	  iat = Exponential(SHORT_IAT*3.0);  /* smaller spikes at half-hour intervals */
	else iat = Exponential(MEAN_IAT/2.0); /* medium background load */
	timer += iat;

      } /* end of for loop */

  } /* end of main */
